/* *********************************************
           Code Example 1
   ********************************************* */

/*
   PROC FCMP

   Create a recursive FCMP function to compute the
   factorial of the value passed in.

   The OUTLIB option is where the FCMP function is stored.
*/
proc fcmp outlib=work.fcmpfunc.fcmp;
   function fcmpfact(x);
     if (x < 0) then return(.);
     else if x eq 0 then return (1);
     else return (x * fcmpfact(x - 1));
   endsub;
run;


/*
   The CMPLIB global option loads the FCMP function library.
*/
options cmplib = work.fcmpfunc;


/* 
   The DATA step calls the FCMP function to compute the factorial.
*/
data _null_;
   /* 
      Compute 5! using the FCMP function.
   */
   result1 = fcmpfact(5);

   /* 
      Compute 5! using the builtin SAS function.
   */
   result2 = fact(5);

   /* 
      Use a put statement to compare results.  5! = 120.
   */
   put "result1: " result1 " result2: " result2;
run;




 
/* *********************************************
           Code Example 2
   ********************************************* */

/* Here is the C code that is compiled in the C library, 
       C:\mathlib\mathfuncs.dll

   The first statement in the C code is the PC compiler option 
   that is necessary to make the function accessible externally.

   The C factorial function is a recursive function that computes
   the factorial of the value passed in.

__declspec(dllexport) double factorial(double x);

double factorial(double x)
{
   if (x < 0) {
      return(0);
   }
   else if (x == 0) {
      return(1);
   }
   else {
      return (x * factorial(x - 1));
   }
}
*/


/* 
   PROC PROTO

   The package statement lists where PROC PROTO information is
   to be stored.  In this example it is work.protofuncs.protopkg.

   The LINK statement loads the C library using the full path
   name, c:\mathlib\mathfuncs.

   The MAPMISS option converts any missing values passed as
   double arguments from a missing value to a -1.

   The prototype of the C factorial function within PROC PROTO
   tells SAS how to call the function.  It lists the parameters
   of the function as well as the return type of the function.
*/
proc proto package = work.protofuncs.protopkg;
   link "C:\mathlib\mathfuncs";
   mapmiss double = -1; 
   double factorial(double num);
run;


/*
   PROC FCMP

   An FCMP function called fcmpfact is created to call the C library 
   factorial function.

   The INLIB option points to where the PROC PROTO package information 
   is stored.

   The OUTLIB option is where the FCMP function is stored.
*/
proc fcmp inlib = work.protofuncs outlib=work.fcmpfuncs.proto;
   function fcmpfact(num);
      y = factorial(num);
      return(y);
   endsub;
run;


/*
   The CMPLIB global option is used to load both the PROTO library and 
   the FCMP function library
*/
options cmplib=(work.protofuncs work.fcmpfuncs);


/*
   The DATA step calls the FCMP factorial function which accesses
   the C library function
*/
data _null_;
   /* 
      Compute 5! using the FCMP/C function.
   */
   result1 = fcmpfact(5);

   /* 
      Compute 5! using the builtin SAS function.
   */
   result2 = fact(5);

   /* 
      Use a put statement to compare results.  5! = 120.
   */
   put "result1: " result1 " result2: " result2;
run;





/* *********************************************
           Code Example 3
   ********************************************* */

/* 
   PROC PROTO

   Simple inline C functions can be declared in PROC PROTO.  A C
   function is created that sets members of a two dimensional
   array to one when the x and y indices are the same
   and to zero when they are not.

   The package statement lists where the information for 
   PROC PROTO is stored.  In this example it is work.protoex.test.

   The prototype of the C array function is part of the 
   PROC PROTO function.  It lists the parameters of the function as
   well as the return type of the function.

   The EXTERNC keyword is used to list the C function name.

   The entire C function is listed after the EXTERNC keyword.
   In this case, the C function uses nested FOR LOOPS to set the
   two dimensional array values.  

   NOTE:  It is very important to see that SAS stores two dimensional
          arrays as a one dimensional array. 
          Detailed information for two dimensional arrays can be found at:
          http://support.sas.com/kb/51/221.html

   The C function ends with the keyword EXTERNCEND.
*/  
proc proto package=work.protoex.test;
   void arrxy (double ** a, int x, int y);
   externc arrxy;
   void arrxy (double **a, int x, int y)
   {
      int i, j;

      for (i=0; i < x; i++) {
         for (j=0; j < y; j++) {
            if (i==j){
               /*
                  The indexing for the two dimensional array is treated 
                  as a one dimensional array.
               */
               a[0][i*x+j] = 1;
            }
            else {
               /*
                  The indexing for the two dimensional array is treated 
                  as a one dimensional array.
               */
               a[0][i*x+j] = 0;
            }
         }
      }
   };
   externcend;
run;


/*
   PROC FCMP

   In this example, the PROC FCMP code does not create a function, but
   simply calls the PROC PROTO arrxy function.

   The INLIB option points to where the PROC PROTO package information 
   is stored.

   A 2 x 2 array is created.  

   The C function, arrxy that is created by PROC PROTO is called
   to set the values of the array.
*/
proc fcmp inlib=work.protoex;
   array b[2,2];
   call arrxy(b, 2, 2);

   /* 
      The put statement will print the values of the 2 dimensional array.  
      The printed values will be:

      b[1, 1]=1 b[1, 2]=0 b[2, 1]=0 b[2, 2]=1
   */
   put b=;
run;
quit;




 
/* *********************************************
           Code Example 4
   ********************************************* */

/* 
   PROC PROTO

   The package statement lists where PROC PROTO information is
   to be stored.  In this example it is work.funcs.protopkg.
  
   Instead of a function definition or prototype, PROC PROTO will
   be used to create a new data type, a C structure.  This structure
   can then be used in PROC FCMP code.
*/  
   
proc proto package = work.funcs.protopkg;
   struct range {
      double high;
      long * low;
   };
run;


/*
   PROC FCMP

   Now PROC FCMP uses the structure definition created in PROC PROTO.
   The fields of the structure are filled in and displayed.

   The INLIB option points to where the PROC PROTO package information 
   is stored.  

   The EXTERNC keyword must be used when creating a variable that is 
   declared to be a C type.  In this case, the variable low is of type long.
*/
proc fcmp inlib = work.funcs;
   struct range   price;
   externc long   low;

   /* 
      PUT statements in FCMP can display structure information.  
      The fields for the structure are initialized to 0 and NULL.
      Expected output: price {high=0, low=NULL}     
   */
   put price=;

   price.high = 1000.00;
   low = 100;

   /* 
      The helper function ISNULL is used to check if the pointer variable
      price.low is set to NULL.
   */
   if (isnull(price.low)) then price.low = low;  

   /* 
     Expected output: price {high=1000, low=100}
   */
   put price=; 
run;




 
/* *********************************************
           Code Example 5
   ********************************************* */

/* 
   PROC PROTO

   The package statement lists where PROC PROTO information is
   to be stored.  In this example it is work.struct.protopkg.
  
   PROC PROTO creates a new data type, a C structure.  This structure
   can then be used in PROC FCMP code.
*/  
proc proto package = work.struct.protopkg;
   struct plot {
      short x;
      short y;
   };
run;


/*
   PROC FCMP

   PROC FCMP uses the structure definition created in PROC PROTO.
   An array of structures is created as well as a single structure element.
   The fields of the structure are filled in and displayed.

   The INLIB option points to where the PROC PROTO package information 
   is stored.  

   The helper routine, STRUCTINDEX, is used to get the element
   from the array at the index
*/
proc fcmp inlib = work.struct;
   struct plot   line[5];
   struct plot   point;

   /* 
      Loop through the array of structures.
      Get the array element at the index using the STRUCTINDEX helper routine. 
      Print the values of each array element.
      The fields for the structure are initialized to 0.
      Expected output for all 5 array elements: point {x=0, y=0}     
   */
   do i = 1 to 5;
      call STRUCTINDEX( line, i, point );
      put point=;
   end;

   /* 
      Loop through the array of structures.
      Get the array element at the index using the STRUCTINDEX helper routine.
      Set the values in each array element.
   */
   do i = 1 to 5;
      call STRUCTINDEX( line, i, point );
      point.x = i;
      point.y = i + 1;
   end;

   /* 
      Loop through the array of structures.
      Get the array element at the index using the STRUCTINDEX helper routine.
      Expected output for the array is:
         point {x=1, y=2}
         point {x=2, y=3}
         point {x=3, y=4}
         point {x=4, y=5}
         point {x=5, y=6}
   */
   do i = 1 to 5;
      call STRUCTINDEX( line, i, point );
      put point=;
   end;
run;




 
/* *********************************************
           Code Example 6
   ********************************************* */

/* 
   PROC PROTO

   The package statement lists where PROC PROTO information is
   to be stored.  In this example it is work.enums.protopkg.
  
   PROC PROTO creates an enumerated type and a C structure.  
   The enumerated list contains the months starting with Jan set to 1.
   The structure uses the enumerated type in one of its fields.
*/  
   
proc proto packet=work.enums.protopkg;
    typedef enum {
        Jan = 1,Feb,Mar,Apr,May,June,July,Aug,Sept,Oct,Nov,Dec
    } Months;
   typedef struct {
      short  day;
      Months month;
      short  year;
   } dateType;
run;


/*
   PROC FCMP

   PROC FCMP uses the structure definition created in PROC PROTO.
   The fields of the structure are filled in and displayed.

   The INLIB option points to where the PROC PROTO package information 
   is stored.  

   NOTE:  When the structure variable is set to the 
          enumerated value, an ampersand must be used before it.
*/
proc fcmp inlib = work.enums;
   struct dateType   todaysdate;

   todaysdate.day = 29;
   todaysdate.month = &Apr;
   todaysdate.year = 2015;
   /* 
     Expected output: todaysdate {day=29, month=4, year=2015}
   */
   put todaysdate=;
run;





/* *********************************************
           Code Example 7
   ********************************************* */

/* 
   PROC PROTO

   A simple inline C function is defined in PROC PROTO to return the
   constant value of VERSION.

   The package statement lists where the PROC PROTO information is 
   to be stored.  In this example it is work.protofunc.ver1.

   A #define constant called VERSION is created.

   The inline function getversion1 will simply return the value of VERSION.
*/  
proc proto package = work.protofunc.ver1;
   #define VERSION 1;
   int getversion1(void);
   externc getversion1;
      int getversion1(void) {
         return VERSION;
      }
   externcend;
run;


/* 
   PROC PROTO

   A second PROC PROTO is declared.  In this example, a simple inline C
   function is declared to return the constant value of VERSION.  

   The package statement lists where the PROC PROTO information is
   to be stored.  In this example it is work.protofunc.ver2.

   A #define constant called VERSION is created.

   NOTE:  It is very important to see that this function uses the
          same name for the constant. 
*/  
proc proto package=work.protofunc.ver2;
   #define VERSION 2;
   int getversion2(void);
   externc getversion2;
      int getversion2(void) {
         return VERSION;
      }
   externcend;
run;


/*
   PROC FCMP

   PROC FCMP now makes use of the PROTO function getversion1 and prints
   prints the return value.

   The INLIB option points to where the PROC PROTO package information 
   is stored.  In this example, work.protofunc will be loaded.

   NOTE:  It is very important to see that the INLIB option points to
          the library and dataset that will be loaded. Both PROC PROTO
          packages are stored in the same library and dataset, so both
          packages will be loaded. 

   PROC FCMP calls the getversion1 function, but it will be the value of
   VERSION from getversion2 that will be used.  This is caused by a naming
   conflict.  The value of VERSION for getversion2 is loaded after 
   getversion1 and overwrites it.
*/
proc fcmp inlib = work.protofunc;

   /* 
      The function getversion1 is called to get the value of VERSION.
   */
   ver = getversion1();

   /* 
      Expected output: ver=2
   */
   put ver=;
run;


